var DirectMail = DirectMail || {};

DirectMail.refreshChartWithIdentifier = function(identifier) {
	var $chart    = $('#' + identifier);
	var chartType = $chart.attr('data-chart-type');
    
    $chart.find('svg').remove();
    
	({
		pie: this.refreshPieChartWithIdentifier,
		bar: this.refreshBarChartWithIdentifier,
		line: this.refreshLineChartWithIdentifier
	}[chartType])(identifier);    
    window.DirectMail_nativeController.didRefreshChart();
};

DirectMail.refreshPieChartWithIdentifier = function(identifier) {
	var $chart = $('#' + identifier);

	(function (raphael) {
		var historyChart = window.DirectMail_nativeController.chartWithIdentifier( identifier );
		var seriess 	 = historyChart.seriess;
		var colors		 = historyChart.colors;
		var chartW		 = $chart.innerWidth();
		var chartH		 = $chart.innerHeight();
		var s			 = Math.min(chartW, chartH);
		var center		 = Math.floor(s/2);

	    raphael(identifier, chartW, chartH).pieChart(
			Math.floor(chartW/2), 
			Math.floor(chartH/2), 
			Math.floor(s * 0.5 * 0.7),
			seriess, 
			colors, 
			'white', 
			historyChart.labelDisplayMode(),
			historyChart.isClickable() ? function (event, series, label) {
				historyChart.didClickLabelInSeries(label, series);
			} : null,
			function (event, series, label) {
				historyChart.didMouseOverLabelInSeries(label, series);
			},
			function (event, series, label) {
				historyChart.didMouseOutLabelInSeries(label, series);
			},
			historyChart.dataUnavailableNotice()
		);
	})(Raphael);
};

DirectMail.refreshLineChartWithIdentifier = function(identifier) {
	var $chart = $('#' + identifier);
	
	(function (raphael) {
		var historyChart = window.DirectMail_nativeController.chartWithIdentifier( identifier );
		var seriess 	 = historyChart.seriess;
		var masterLabels = historyChart.masterLabels;
		var colors		 = historyChart.colors;
		var sizingChart  = raphael(0, 0, 0, 0);
		var metrics      = sizingChart.lineChartMetrics(seriess, masterLabels, $chart.data("force-width"));
		var chartW	     = metrics.chartElementWidth;
		
		$chart.width(chartW);
		sizingChart.remove();
		
		var chartH = $chart.innerHeight();
		
		raphael(identifier, chartW, chartH).lineChart(
			metrics,
			chartH,
			seriess, 
			masterLabels,
			colors,
			function (event, series, label) {
				historyChart.didMouseOverLabelInSeries(label, series);
			},
			function (event, series, label) {
				historyChart.didMouseOutLabelInSeries(label, series);
			}
		);
	})(Raphael);
};

DirectMail.refreshBarChartWithIdentifier = function(identifier) {
	var $chart = $('#' + identifier);
	
	(function (raphael) {
		var historyChart = window.DirectMail_nativeController.chartWithIdentifier( identifier );
		var seriess 	 = historyChart.seriess;
		var masterLabels = historyChart.masterLabels;
		var colors		 = historyChart.colors;
		var sizingChart  = raphael(0, 0, 0, 0);
		var metrics      = sizingChart.barChartMetrics(seriess, masterLabels);
		var chartW	     = metrics.chartElementWidth;
		
		$chart.width(chartW);
		sizingChart.remove();
		
        var chartH = $chart.innerHeight();
		
		raphael(identifier, chartW, chartH).barChart(
			metrics,
			chartH,
			seriess, 
			masterLabels,
			colors,
			historyChart.isClickable() ? function (event, series, label) {
				historyChart.didClickLabelInSeries(label, series);
			} : null,
			function (event, series, label) {
				historyChart.didMouseOverLabelInSeries(label, series);
			},
			function (event, series, label) {
				historyChart.didMouseOutLabelInSeries(label, series);
			}
		);
	})(Raphael);
};

DirectMail.hsbaFromColor = function(color) {
	return {
		h: color.hueComponent(),
		s: color.saturationComponent(),
		b: color.brightnessComponent(),
		a: color.alphaComponent()
	};
};

DirectMail.hsbaDeclarationWithValues = function(hsba) {
	return "hsba(" + hsba.h + "," + hsba.s + "," + hsba.b + "," + hsba.a + ")";
};

DirectMail.systemAppearance = {
    "color-label" : "black",
    "color-label-secondary" : "gray",
    "color-label-tertiary" : "rgb(230, 230, 230)",
    "color-text-background" : "white",
    "color-grid" : "rgba(0, 0, 0, 0.1)"
};

DirectMail.populateSystemAppearanceValues = function() {
    if ( window.DirectMail_nativeController.populateSystemAppearanceValues !== undefined ) {
        window.DirectMail_nativeController.populateSystemAppearanceValues( DirectMail.systemAppearance );
    }
};

DirectMail.systemAppearanceDidChange = function() {
    DirectMail.populateSystemAppearanceValues();
    $('.uses-system-appearance').each(function(index) {
        DirectMail.refreshChartWithIdentifier($(this).attr('id'));
    });
};

DirectMail.resizeCharts = function() {
    $('.fills-parent').each(function(index) {
        DirectMail.refreshChartWithIdentifier($(this).attr('id'));
    });
};

DirectMail.forceWidthForChartWithIdentifier = function(identifier, width) {
    $('#' + identifier).data("force-width", width);
    DirectMail.resizeCharts();
};

$(document).ready(function() {
	$(window).resize(function() {
        DirectMail.resizeCharts();
	});
    DirectMail.populateSystemAppearanceValues();
});
