package Mail::DeliveryStatus::BounceParser;
use 5.006;
use strict;
use warnings;
our$VERSION='1.523';
$VERSION=eval$VERSION;
use MIME::Parser;
use Mail::DeliveryStatus::Report;
use vars qw($EMAIL_ADDR_REGEX);
$EMAIL_ADDR_REGEX=qr{
# Avoid using something like Email::Valid
# Full rfc(2)822 compliance isn't exactly what we want, and this seems to work
# for most real world cases
(?:<|^|\s)            # Space, or the start of a string
([^\s\/<]+            # some non-space, non-/ characters; none are <
\@                    # at sign (duh)
(?:[-\w]+\.)+[-\w]+)  # word characters or hypens organized into
                      # at least two dot-separated words
(?:$|\s|>)            # then the end
}sx;
my$Not_An_Error=qr/
    \b delayed \b
  | \b warning \b
  | transient.{0,20}\serror
  | Your \s message .{0,100} was \s delivered \s to \s the \s following \s recipient
/six;
my$Really_An_Error=qr/this is a permanent error|permanent fatal errors/i;
my$Returned_Message_Below=qr/(
    (?:original|returned) \s message \s (?:follows|below)
  | (?: this \s is \s a \s copy \s of
      | below \s this \s line \s is \s a \s copy
    ) .{0,100} \s message\.?
  | message \s header \s follows
  | ^ (?:return-path|received|from):
)\s+/sixm;
my@Preprocessors=qw(
  p_html_report
  p_ims
  p_aol_senderblock
  p_novell_groupwise
  p_plain_smtp_transcript
  p_xdelivery_status
);
sub parse{my($class,$data,$arg)=@_;
my$parser=new MIME::Parser;
$parser->output_to_core(1);
my$message;
if(not$data){print STDERR"BounceParser: expecting bounce mesage on STDIN\n" if-t STDIN;
$message=$parser->parse(\*STDIN);}elsif(not ref$data){$message=$parser->parse_data($data);}elsif(ref$data eq"ARRAY"){$message=$parser->parse_data($data);}else{$message=$parser->parse($data);}my$self=bless{reports=>[],is_bounce=>1,log=>$arg->{log},parser=>$parser,orig_message_id=>undef,prefer_final_recipient=>$arg->{prefer_final_recipient},},$class;
$self->log("received message with type ".(defined($message->effective_type)?$message->effective_type:"undef").", subject ".(defined($message->head->get("subject"))?$message->head->get("subject"):"CAN'T GET SUBJECT"));
foreach my $preprocessor(@Preprocessors){if(my$newmessage=$self->$preprocessor($message)){$message=$newmessage;}}$self->{message}=$message;
$self->log("now the message is type ".$message->effective_type.", subject ".(defined($message->head->get("subject"))?$message->head->get("subject"):"CAN'T GET SUBJECT"));
my$first_part=_first_non_multi_part($message);
{last unless($message->head->get("x-delivery-agent")and$message->head->get("X-Delivery-Agent")=~/^TMDA/);
$self->log("looks like a challenge/response autoresponse; ignoring.");
$self->{type}="Challenge / Response system autoreply";
$self->{is_bounce}=0;
return$self;}{last unless($message->head->get("X-Bluebottle-Request")and$first_part->stringify_body=~/This account is protected by Bluebottle/);
$self->log("looks like a challenge/response autoresponse; ignoring.");
$self->{type}="Challenge / Response system autoreply";
$self->{is_bounce}=0;
return$self;}{last if$message->effective_type eq 'multipart/report';
last if!$first_part||$first_part->effective_type ne 'text/plain';
my$string=$first_part->as_string;
last if length($string)>3000;
last if$string!~/auto.{0,20}reply|return receipt|vacation|(out|away|on holiday).*office/i;
$self->log("looks like a vacation autoreply, ignoring.");
$self->{type}="vacation autoreply";
$self->{is_bounce}=0;
return$self;}{last if$message->effective_type eq 'multipart/report';
last if!$first_part||$first_part->effective_type ne 'text/plain';
my$string=$first_part->as_string;
last if length($string)>3000;
last if$string!~/(address .{0,60} changed | domain .{0,40} retired) .*
          (has\s*been|was|have|will\s*be) \s* (forwarded|delivered)/six;
$self->log('looks like an address-change autoreply, ignoring');
$self->{type}='informational address-change autoreply';
$self->{is_bounce}=0;
return$self;}if($message->effective_type eq"text/plain" and(length$message->as_string)<3000 and$message->bodyhandle->as_string=~m/norton\sassociates\swebshield|content\s+filter/ix){$self->log("looks like a virus/spam block, ignoring.");
$self->{type}="virus/spam false positive";
$self->{is_bounce}=0;
return$self;}if($message->effective_type eq"text/plain" and$message->bodyhandle->as_string=~/transient.*error/is){$self->log("seems like a nonfatal error, ignoring.");
$self->{is_bounce}=0;
return$self;}{my$part_for_maybe_transient;
$part_for_maybe_transient=$message if$message->effective_type eq"text/plain";
($part_for_maybe_transient)=grep{$_->effective_type eq"text/plain"}$message->parts if$message->effective_type=~/multipart/&&$message->effective_type ne 'multipart/report';
if($part_for_maybe_transient){my$string=$part_for_maybe_transient->bodyhandle->as_string;
my$transient_pos=_match_position($string,$Not_An_Error);
last unless defined$transient_pos;
my$permanent_pos=_match_position($string,$Really_An_Error);
my$orig_msg_pos=_match_position($string,$Returned_Message_Below);
last if _position_before($permanent_pos,$orig_msg_pos);
if(_position_before($transient_pos,$orig_msg_pos)){$self->log("transient error, ignoring.");
$self->{is_bounce}=0;
return$self;}}}if($message->effective_type=~/multipart/){if(my($orig_message)=grep{$_->effective_type eq"message/rfc822"}$message->parts){my$orig_message_id=$orig_message->parts(0)->head->get("message-id");
if($orig_message_id){chomp$orig_message_id;
$self->log("extracted original message-id $orig_message_id from the original rfc822/message");}else{$self->log("Couldn't extract original message-id from the original rfc822/message");}$self->{orig_message_id}=$orig_message_id;
$self->{orig_message}=$orig_message->parts(0);}if(!$self->{orig_message_id}and my($rfc822_headers)=grep{lc$_->effective_type eq"text/rfc822-headers"}$message->parts){my$orig_head=Mail::Header->new($rfc822_headers->body);
chomp($self->{orig_message_id}=$orig_head->get("message-id"));
$self->{orig_header}=$orig_head;
$self->log("extracted original message-id $self->{orig_message_id} from text/rfc822-headers");}}if(!$self->{orig_message_id}){if($message->bodyhandle and$message->bodyhandle->as_string=~/Message-ID: (\S+)/i){$self->{orig_message_id}=$1;
$self->log("found a message-id $self->{orig_message_id} in the body.");}}if(!$self->{orig_message_id}){$self->log("couldn't find original message id.");}if($message->effective_type eq"multipart/report"){my($delivery_status)=grep{$_->effective_type eq"message/delivery-status"}$message->parts;
if(!$delivery_status){my@delivery_status_parts=grep{$_->effective_type=~m{text/plain} and not$_->is_multipart}$message->parts;
push@{$self->{reports}},$self->_extract_reports(@delivery_status_parts);}else{my%global=("reporting-mta"=>undef,"arrival-date"=>undef);
my($seen_action_expanded,$seen_action_failed);
my$delivery_status_body=eval{$delivery_status->bodyhandle->as_string}||'';
foreach my $para(split/\n{2,}/,$delivery_status_body){$para=~s/\A\n+//g;
$para=~s/\r/ /g;
my$report=Mail::Header->new([split/\n/,$para]);
$report->unfold;
if(my$action=lc$report->get('Action')){$action=~s/^\s+//;
if($action=~s/^\s*([a-z]+)\b.*/$1/s){if($action eq 'expanded'){$seen_action_expanded=1;}elsif($action eq 'failed'){$seen_action_failed=1;}else{$self->log("message/delivery-status says 'Action: \L$1'");
$self->{type}='delivery-status \L$1';
$self->{is_bounce}=0;
return$self;}}}for my $hdr(qw(Reporting-MTA Arrival-Date)){my$val=$global{$hdr}||=$report->get($hdr);
if(defined($val)){$report->replace($hdr=>$val)}}my$email;
if($self->{prefer_final_recipient}){$email=$report->get("final-recipient")||$report->get("original-recipient");}else{$email=$report->get("original-recipient")||$report->get("final-recipient");}next unless$email;
my$reason=$report->get("diagnostic-code");
$email=~s/[^;]+;\s*//;
if(defined$reason){$reason=~s/[^;]+;\s*//;}$email=_cleanup_email($email);
$report->replace(email=>$email);
if(defined$reason){$report->replace(reason=>$reason);}else{$report->delete("reason");}if(my$status=$report->get('Status')){if($status=~/^5\.1\.[01]$/){$report->replace(std_reason=>"user_unknown");}elsif($status eq"5.1.2"){$report->replace(std_reason=>"domain_error");}elsif($status eq"5.2.2"){$report->replace(std_reason=>"over_quota");}else{$report->replace(std_reason=>_std_reason($report->get("diagnostic-code")));}}else{$report->replace(std_reason=>_std_reason($report->get("diagnostic-code")));}my$diag_code=$report->get("diagnostic-code");
my$host;
if(defined$diag_code){($host)=$diag_code=~/\bhost\s+(\S+)/;}$report->replace(host=>($host))if$host;
my($code);
if(defined$diag_code){($code)=$diag_code=~m/ ( ( [245] \d{2} ) \s | \s ( [245] \d{2} ) (?!\.) ) /x;}if($code){$report->replace(smtp_code=>$code);}if(not$report->get("host")){my$email=$report->get("email");
if(defined$email){my$host=($email=~/\@(.+)/)[0];
$report->replace(host=>$host)if$host;}}if($report->get("smtp_code")and($report->get("smtp_code")=~/^2../)){$self->log("smtp code is ".$report->get("smtp_code")."; no_problemo.");}unless($arg->{report_non_bounces}){if($report->get("std_reason")eq"no_problemo"){$self->log("not actually a bounce: ".$report->get("diagnostic-code"));
next;}}push@{$self->{reports}},Mail::DeliveryStatus::Report->new([split/\n/,$report->as_string]);}if($seen_action_expanded&&!$seen_action_failed){$self->log(q[message/delivery-status says 'Action: expanded']);
$self->{type}='delivery-status expanded';
$self->{is_bounce}=0;
return$self;}}}elsif($message->effective_type=~/multipart/){my@delivery_status_parts=grep{$_->effective_type=~m{text/plain} and not$_->is_multipart}$message->parts;
push@{$self->{reports}},$self->_extract_reports(@delivery_status_parts);}elsif($message->effective_type=~m{text/plain}){my$body_string=$message->bodyhandle->as_string||'';
if($body_string=~$Returned_Message_Below){my($stuff_before,$stuff_splitted,$stuff_after)=split$Returned_Message_Below,$message->bodyhandle->as_string,2;
push@{$self->{reports}},$self->_extract_reports($stuff_before);
$self->{orig_text}=$stuff_after;}elsif($body_string=~/(.+)\n\n(.+?Message-ID:.+)/is){push@{$self->{reports}},$self->_extract_reports($1);
$self->{orig_text}=$2;}else{push@{$self->{reports}},$self->_extract_reports($body_string);
$self->{orig_text}=$body_string;}}return$self;}BEGIN{*new=\&parse}sub log{my($self,@log)=@_;
if(ref$self->{log}eq"CODE"){$self->{log}->(@_);}return 1;}sub _extract_reports{my$self=shift;
if(@_>1){return map{$self->_extract_reports($_)}@_}my$text=shift;
$text=$text->bodyhandle->as_string if ref$text;
my%by_email;
return unless$text;
my@split=split($EMAIL_ADDR_REGEX,$text);
foreach my $i(0..$#split){next if$i%2==0;
my$email=_cleanup_email($split[$i]);
if($split[$i-1]=~/they are not accepting mail from/){next;}my$std_reason="unknown";
$std_reason=_std_reason($split[$i+1])if$#split>$i;
$std_reason=_std_reason($split[$i-1])if$std_reason eq"unknown";
next if(exists$by_email{$email}and$by_email{$email}->{std_reason}ne"unknown" and$std_reason eq"unknown");
my$reason=$split[$i-1];
$reason=~s/(.*?). (Your mail to the following recipients could not be delivered)/$2/;
$by_email{$email}={email=>$email,raw=>join("",@split[$i-1..$i+1]),std_reason=>$std_reason,reason=>$reason};}my@toreturn;
foreach my $email(keys%by_email){my$report=Mail::DeliveryStatus::Report->new();
$report->header_hashref($by_email{$email});
push@toreturn,$report;}return@toreturn;}sub is_bounce{return shift->{is_bounce};}sub reports{return@{shift->{reports}}}sub addresses{return map{$_->get("email")}shift->reports;}sub orig_message_id{return shift->{orig_message_id};}sub orig_message{return shift->{orig_message}}sub orig_header{return shift->{orig_header}}sub orig_text{return shift->{orig_text}}sub _std_reason{local$_=shift;
if(!defined$_){return"unknown";}if(/(?:domain|host|service)\s+(?:not\s+found|unknown|not\s+known)/i or/couldn't\s+find\s+any\s+host/i){return"domain_error"}if(/try.again.later/is or/mailbox\b.*\bfull/ or/storage/i          or/quota/i            or/\s552\s/           or/\s#?5\.2\.2\s/     or/User\s+mailbox\s+exceeds\s+allowed\s+size/){return"over_quota";}my$user_re=qr'(?: mailbox  | user | recipient | address (?: ee)?
       | customer | account | e-?mail | <? $EMAIL_ADDR_REGEX >? )'ix;
if(/\s \(? \#? 5\.1\.[01] \)? \s/x or/$user_re\s+(?:\S+\s+)? (?:is\s+)?                  # Generic
     (?: (?: un|not\s+) (?: known | recognized )
      | [dw]oes\s?n[o']?t 
     (?: exist|found ) | disabled | expired ) /ix or/no\s+(?:such)\s+?$user_re/i or/unrouteable address/i or/inactive user/i or/unknown local part/i or/user\s+doesn't\s+have\s+a/i or/account\s+has\s+been\s+(?:disabled|suspended)/i or/$user_re\s+(?:suspended|discontinued)/i or/unknown\s+$user_re/i or/$user_re\s+(?:is\s+)?(?:inactive|unavailable)/i or/(?:(?:in|not\s+a\s+)?valid|no such)\s$user_re/i or/$user_re\s+(?:was\s+)?not\s+found/i or/$user_re \s+ (?:is\s+)? (?:currently\s+)?          # ATT, generic
     (?:suspended|unavailable)/ix or/address is administratively disabled/i or/no $user_re\s+(?:here\s+)?by that name/i or/<?$EMAIL_ADDR_REGEX>? is invalid/i or/address.*not known here/i or/recipient\s+(?:address\s+)?rejected/i or/not\s+listed\s+in\s+Domino/i or/account not activated/i or/not\s+our\s+customer/i or/doesn't handle mail for that user/i){return"user_unknown";}if(/domain\s+syntax/i or/timed\s+out/i or/route\s+to\s+host/i or/connection\s+refused/i or/no\s+data\s+record\s+of\s+requested\s+type/i or/Malformed name server reply/i or/as\s+a\s+relay,\s+but\s+I\s+have\s+not\s+been\s+configured\s+to\s+let/i or/550\s+relay\s+not\s+permitted/i or/550\s+relaying\s+denied/i or/Relay\s+access\s+denied/i or/Relaying\s+denied/i){return"domain_error";}if(/Blocked\s+by\s+SpamAssassin/i or/spam\s+rejection/i or/identified\s+SPAM,\s+message\s+permanently\s+rejected/i or/Mail\s+appears\s+to\s+be\s+unsolicited/i or/Message\s+rejected\s+as\s+spam\s+by\s+Content\s+Filtering/i){return"spam";}return"unknown";}sub p_html_report{my($self,$message)=@_;
if($message->effective_type eq"multipart/mixed"){my@keep=grep{$_->effective_type eq"multipart/report"}$message->parts;
if(@keep){my@new_parts=$keep[0]->parts;
$message->parts(\@new_parts);
$message->effective_type("multipart/report");}}return$message;}sub p_ims{my$self=shift;
my$message=shift;
return unless($message->head->get("X-Mailer")||'')=~/Internet Mail Service/i;
if($message->is_multipart){return unless my($error_part)=grep{$_->effective_type eq"text/plain"}$message->parts;
return unless my($actual_error)=$error_part->as_string=~/did not reach the following recipient\S+\s*(.*)/is;
if(my$io=$error_part->open("w")){$io->print($actual_error);
$io->close;}}else{return unless my($actual_error)=$message->bodyhandle->as_string=~/did not reach the following recipient\S+\s*(.*)/is;
my($stuff_before,$stuff_after)=split/^(?=Message-ID:|Received:)/m,$message->bodyhandle->as_string;
$stuff_before=~s/.*did not reach the following recipient.*?$//ism;
$self->log("rewrote IMS into plain/report.");
return$self->new_plain_report($message,$stuff_before,$stuff_after);}return$message;}sub p_aol_senderblock{my$self=shift;
my$message=shift;
return unless($message->head->get("Mailer")||'')=~/AirMail/i;
return unless$message->effective_type eq"text/plain";
return unless$message->bodyhandle->as_string=~/Your mail to the following recipients could not be delivered because they are not accepting mail/i;
my($host)=$message->head->get("From")=~/\@(\S+)>/;
my$rejector;
my@new_output;
for(split/\n/,$message->bodyhandle->as_string){s/Sorry \S+?@\S+?\.//g;
if(/because they are not accepting mail from (\S+?):?/i){$rejector=$1;
push@new_output,$_;
next;}if(/^\s*(\S+)\s*$/){my$recipient=$1;
if($recipient=~/\@/){push@new_output,$_;
next;}s/^(\s*)(\S+)(\s*)$/$1$2\@$host$3/;
push@new_output,$_;
next;}push@new_output,$_;
next;}push@new_output,("# rewritten by BounceParser: p_aol_senderblock()","");
if(my$io=$message->open("w")){$io->print(join"\n",@new_output);
$io->close;}return$message;}sub p_novell_groupwise{my$self=shift;
my$message=shift;
return unless($message->head->get("X-Mailer")||'')=~/Novell Groupwise/i;
return unless$message->effective_type eq"multipart/mixed";
return unless my($error_part)=grep{$_->effective_type eq"text/plain"}$message->parts;
my($host)=$message->head->get("From")=~/\@(\S+)>?/;
my@new_output;
for(split/\n/,$error_part->bodyhandle->as_string){if(/^(\s*)(\S+)(\s+\(.*\))$/){my($space,$recipient,$reason)=($1,$2,$3);
if($recipient=~/\@/){push@new_output,$_;
next;}$_=join"",$space,"$2\@$host",$reason;
push@new_output,$_;next;}push@new_output,$_;next;}push@new_output,("# rewritten by BounceParser: p_novell_groupwise()","");
if(my$io=$error_part->open("w")){$io->print(join"\n",@new_output);
$io->close;}return$message;}sub p_plain_smtp_transcript{my($self,$message)=(shift,shift);
return unless$message->effective_type eq"text/plain";
return unless$message->bodyhandle->as_string=~/The following addresses had permanent fatal errors/;
return unless$message->bodyhandle->as_string=~/Transcript of session follows/;
return unless$message->bodyhandle->as_string=~/Message .* follows/;
my($stuff_before,$stuff_after)=split/^.*Message (?:header|body) follows.*$/im,$message->bodyhandle->as_string,2;
my%by_email=$self->_analyze_smtp_transcripts($stuff_before);
my@paras=_construct_delivery_status_paras(\%by_email);
my@new_output;
my($reporting_mta)=_cleanup_email($message->head->get("From"))=~/\@(\S+)/;
chomp(my$arrival_date=$message->head->get("Date"));
push@new_output,"Reporting-MTA: $reporting_mta" if$reporting_mta;
push@new_output,"Arrival-Date: $arrival_date" if$arrival_date;
push@new_output,"";
push@new_output,map{@$_,""}@paras;
return$self->new_multipart_report($message,$stuff_before,join("\n",@new_output),$stuff_after);}sub _construct_delivery_status_paras{my%by_email=%{shift()};
my@new_output;
foreach my $email(sort keys%by_email){push@new_output,["Final-Recipient: RFC822; $email","Action: failed","Status: 5.0.0",($by_email{$email}->{host}?("Remote-MTA: DNS; $by_email{$email}->{host}"):()),_construct_diagnostic_code(\%by_email,$email),];}return@new_output;}sub _construct_diagnostic_code{my%by_email=%{shift()};
my$email=shift;
join(" ","Diagnostic-Code: X-BounceParser;",($by_email{$email}->{'host'}?"host $by_email{$email}->{'host'} said:":()),($by_email{$email}->{'smtp_code'}),(join", ",@{$by_email{$email}->{'errors'}}));}sub _analyze_smtp_transcripts{my$self=shift;
my$plain_smtp_transcript=shift;
my(%by_email,$email,$smtp_code,@error_strings,$host);
for(split/\n\n|(?=>>>)/,$plain_smtp_transcript){$email=_cleanup_email($1)if/RCPT TO:\s*(\S+)/im;
$by_email{$email}->{host}=$host if$email;
if(/while talking to (\S+)/im){$host=$1;
$host=~s/[.:;]+$//g;}if(/<<< (\d\d\d) (.*)/m){$by_email{$email}->{smtp_code}=$1;
push@{$by_email{$email}->{errors}},$2;}if(/^(\d\d\d)\b.*(<\S+\@\S+>)\.*\s+(.+)/m){$email=_cleanup_email($2);
$by_email{$email}->{smtp_code}=$1;
push@{$by_email{$email}->{errors}},$3;}}delete$by_email{''};
return%by_email;}sub new_plain_report{my($self,$message,$error_text,$orig_message)=@_;
$orig_message=~s/^\s+//;
my$newmessage=$message->dup();
$newmessage->make_multipart("plain-report");
$newmessage->parts([]);
$newmessage->attach(Type=>"text/plain",Data=>$error_text);
my$orig_message_mime=MIME::Entity->build(Type=>"multipart/transitory");
$orig_message_mime->add_part($self->{parser}->parse_data($orig_message));
$orig_message_mime->head->mime_attr("content-type"=>"message/rfc822");
$newmessage->add_part($orig_message_mime);
$self->log("created new plain-report message.");
return$newmessage;}sub new_multipart_report{my($self,$message,$error_text,$delivery_status,$orig_message)=@_;
$orig_message=~s/^\s+//;
my$newmessage=$message->dup();
$newmessage->make_multipart("report");
$newmessage->parts([]);
$newmessage->attach(Type=>"text/plain",Data=>$error_text);
$newmessage->attach(Type=>"message/delivery-status",Data=>$delivery_status);
my$orig_message_mime=MIME::Entity->build(Type=>"multipart/transitory",Top=>0);
$orig_message_mime->add_part($self->{parser}->parse_data($orig_message));
$orig_message_mime->head->mime_attr("content-type"=>"message/rfc822");
$newmessage->add_part($orig_message_mime);
$self->log("created new multipart-report message.");
return$newmessage;}sub _cleanup_email{my$email=shift;
for($email){chomp;
tr/[()]//d;
s/^To:\s*//i;
s/[.:;]+$//;
s/<(.+)>/$1/;
s/.*:SMTP=//;
s/^\s+//;
s/\s+$//;
s/\.{3}\S+//;
s/^SMTP://;}return$email;}sub p_xdelivery_status{my($self,$message)=@_;
for($message->parts_DFS){$_->effective_type('message/delivery-status')if$_->effective_type eq 'message/xdelivery-status';}}sub _first_non_multi_part{my($entity)=@_;
my$part=$entity;
$part=$part->parts(0)or return while$part->is_multipart;
return$part;}sub _position_before{my($pos_a,$pos_b)=@_;
return 1 if defined($pos_a)&&(!defined($pos_b)||$pos_a<$pos_b);
return;}sub _match_position{my($string,$regex)=@_;
return$string=~$regex?$-[0]:undef;}1;
